﻿/*
 * Copyright (C)  2011  Axel Kesseler
 * 
 * This software is free and you can use it for any purpose. Furthermore, 
 * you are free to copy, to modify and/or to redistribute this software.
 * 
 * In addition, this software is distributed in the hope that it will be 
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

using System;
using System.IO;
using System.Text;
using System.Drawing;
using System.Reflection;
using System.Diagnostics;
using System.Windows.Forms;
using System.Collections.Generic;
using System.Runtime.InteropServices;

using plexdata.Controls;
using plexdata.Utilities.Clipboard;

namespace plexdata.ClipboardViewer
{
    public partial class MainForm : Form
    {
        public event EventHandler<EventArgs> ClipboardChanged;

        private class NativeClipboardItem
        {
            public NativeClipboardItem(int format, int size, string name, string type, byte[] data)
                : base()
            {
                this.Format = format;
                this.Size = size;
                this.Name = name;
                this.Type = type;
                this.Data = data;
            }

            public int Format { get; private set; }
            public int Size { get; private set; }
            public string Name { get; private set; }
            public string Type { get; private set; }
            public byte[] Data { get; private set; }

            public override string ToString()
            {
                if (String.IsNullOrEmpty(this.Name))
                {
                    return base.ToString();
                }
                else
                {
                    return this.Name;
                }
            }
        }

        private IntPtr hNextViewer = IntPtr.Zero;

        private Settigns settigns = new Settigns();

        public MainForm()
        {
            this.InitializeComponent();

            this.Text = AboutBox.Title;
            this.Icon = Icon.FromHandle(plexdata.ClipboardViewer.Properties.Resources.MainIcon.GetHicon());
        }

        protected override void WndProc(ref Message message)
        {
            switch (message.Msg)
            {
                case WM_CREATE:

                    // Add window to the clipboard viewer chain. 
                    this.hNextViewer = SetClipboardViewer(message.HWnd);
                    break;

                case WM_DESTROY:

                    // Remove window from the clipboard viewer chain. 
                    ChangeClipboardChain(message.HWnd, this.hNextViewer);
                    break;

                case WM_CHANGECBCHAIN:

                    // If next window is closing then repair the chain. 
                    if (message.WParam == this.hNextViewer)
                    {
                        this.hNextViewer = message.LParam;
                    }
                    // Otherwise, pass the message to the next link. 
                    else if (this.hNextViewer != IntPtr.Zero)
                    {
                        SendMessage(this.hNextViewer, message.Msg, message.WParam, message.LParam);
                    }
                    break;

                case WM_DRAWCLIPBOARD:

                    // Handle changes in clipboard contents. 
                    this.OnClipboardChanged();
                    // Pass this message to the next window in clipboard viewer chain. 
                    SendMessage(this.hNextViewer, message.Msg, message.WParam, message.LParam);
                    break;

                default:

                    // Nothing else to do.
                    break;
            }
            base.WndProc(ref message);
        }

        private void OnClipboardChanged()
        {
            Cursor oldCursor = this.Cursor;
            this.Cursor = Cursors.WaitCursor;
            try
            {
                // Reset all visible information.
                this.lblSelectionInfo.Text = "";
                this.hexViewer.Buffer = null;

                // Disable toolbar buttons.
                this.tbbSave.Enabled = false;
                this.tbbEmpty.Enabled = false;
                this.tbbText.Enabled = false;
                this.tbbImage.Enabled = false;

                // Clear old items.
                this.lstSelector.Items.Clear();

                // Reinitialize list content with new data.
                using (NativeClipboard cb = new NativeClipboard(this))
                {
                    // Safety check!
                    if (cb.IsOpened())
                    {
                        foreach (int format in cb.GetFormats())
                        {
                            this.lstSelector.Items.Add(new NativeClipboardItem(
                                format,
                                cb.GetSize(format),
                                cb.GetName(format),
                                cb.GetName(format, false),
                                cb.GetData(format)));
                        }
                    }
                }

                // Select first entry if possible.
                if (this.lstSelector.Items.Count > 0)
                {
                    this.lstSelector.SelectedIndex = 0;
                    this.tbbEmpty.Enabled = true;
                }

                // Raise event if possible.
                if (ClipboardChanged != null)
                {
                    ClipboardChanged(this, EventArgs.Empty);
                }
            }
            catch (Exception exception)
            {
                Debug.WriteLine(exception);
            }
            finally
            {
                this.Cursor = oldCursor;
            }
        }

        #region Clipboard viewer related imports.

        [DllImport("user32.dll", SetLastError = true)]
        private static extern IntPtr SetClipboardViewer(IntPtr hWndNewViewer);

        [DllImport("user32.dll", SetLastError = true)]
        private static extern bool ChangeClipboardChain(IntPtr hWndRemove, IntPtr hWndNewNext);

        [DllImport("user32.dll", SetLastError = true)]
        private static extern IntPtr SendMessage(IntPtr hWnd, int message, IntPtr wParam, IntPtr lParam);

        #endregion // Clipboard viewer related imports.

        #region Clipboard viewer related constant definitions.

        private const int WM_CREATE = 0x0001;
        private const int WM_DESTROY = 0x0002;

        private const int WM_CHANGECBCHAIN = 0x030D;
        private const int WM_DRAWCLIPBOARD = 0x0308;

        #endregion // Clipboard viewer related constant definitions.

        #region Form event handler implementation.

        private void MainForm_Move(object sender, EventArgs args)
        {
            if (this.WindowState == FormWindowState.Normal)
            {
                this.settigns.General.Location = this.Location;
            }
        }

        private void MainForm_Resize(object sender, EventArgs args)
        {
            if (this.WindowState == FormWindowState.Normal)
            {
                this.settigns.General.Size = this.Size;
            }
        }

        private void MainForm_Load(object sender, EventArgs args)
        {
            Settigns settigns;
            if (Settigns.Load(Settigns.Filename, out settigns))
            {
                this.settigns = (Settigns)settigns.Clone();
                this.ApplySettings(settigns);
            }
        }

        private void MainForm_FormClosing(object sender, FormClosingEventArgs args)
        {
            Settigns.Save(Settigns.Filename, this.settigns);
        }

        #endregion // Form event handler implementation.

        #region Toolbar button event handler implementation.

        private void tbbExit_Click(object sender, EventArgs args)
        {
            this.Close();
        }

        private void tbbSave_Click(object sender, EventArgs args)
        {
            SaveFileDialog dialog = new SaveFileDialog();
            dialog.Filter = "Binary Files (*.bin)|*.bin|All Files (*.*)|*.*";
            dialog.InitialDirectory = Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location);
            dialog.RestoreDirectory = true;

            if (DialogResult.OK == dialog.ShowDialog(this))
            {
                try
                {
                    NativeClipboardItem selected = (this.lstSelector.SelectedItem as NativeClipboardItem);
                    if (selected != null)
                    {
                        using (FileStream file = new FileStream(dialog.FileName, FileMode.CreateNew))
                        {
                            file.Write(selected.Data, 0, selected.Data.Length);
                        }
                    }
                }
                catch (Exception exception)
                {
                    MessageBox.Show(
                        this, "File couldn't be saved!\n\n" + exception.ToString(),
                        "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
        }

        private void tbbEmpty_Click(object sender, EventArgs args)
        {
            using (NativeClipboard cb = new NativeClipboard(this))
            {
                if (cb.IsOpened()) { cb.Empty(); }
            }
        }
        
        private void tbbText_Click(object sender, EventArgs args)
        {
            TextViewDialog dialog = new TextViewDialog(this);
            dialog.Size = this.settigns.TextView.Size;
            dialog.Location = new Point(this.Location.X + 25, this.Location.Y + 65);
            dialog.ShowDialog(this);
            this.settigns.TextView.Size = dialog.Size;
        }

        private void tbbImage_Click(object sender, EventArgs args)
        {
            ImageViewDialog dialog = new ImageViewDialog(this);
            dialog.Size = this.settigns.ImageView.Size;
            dialog.Location = new Point(this.Location.X + 25, this.Location.Y + 65);
            dialog.ShowDialog(this);
            this.settigns.ImageView.Size = dialog.Size;
        }
        
        private void tbbSettings_Click(object sender, EventArgs args)
        {
            SettingsDialog dialog = new SettingsDialog(this, this.settigns);
            if (DialogResult.OK == dialog.ShowDialog(this))
            {
                this.settigns = (Settigns)dialog.Settigns.Clone();
            }
            this.ApplySettings(this.settigns);
        }

        private void tbbAbout_Click(object sender, EventArgs args)
        {
            AboutBox about = new AboutBox();
            about.ShowDialog(this);
        }

        #endregion // Toolbar button event handler implementation.

        private void SelectorList_SelectedIndexChanged(object sender, EventArgs args)
        {
            Cursor oldCursor = this.Cursor;
            this.Cursor = Cursors.WaitCursor;
            try
            {
                NativeClipboardItem selected = (this.lstSelector.SelectedItem as NativeClipboardItem);
                if (selected != null)
                {
                    // Set status label.
                    this.lblSelectionInfo.Text = String.Format(
                        "Name: {0}, Type: {1}, Format: #{2}, Size: #{3}",
                        selected.Name,
                        selected.Type,
                        selected.Format,
                        selected.Size);

                    // Set binary data.
                    this.hexViewer.Buffer = selected.Data;

                    // Enable depending toolbar buttons.
                    this.tbbSave.Enabled = (selected.Data != null && selected.Data.Length > 0);
                    this.tbbText.Enabled = Clipboard.ContainsText();
                    this.tbbImage.Enabled = Clipboard.ContainsImage();
                }
            }
            catch (Exception exception)
            {
                Debug.WriteLine(exception);
            }
            finally
            {
                this.Cursor = oldCursor;
            }
        }

        public void ApplySettings(Settigns settigns)
        {
            this.Location = settigns.General.Location;
            this.Size = settigns.General.Size;

            this.lstSelector.ForeColor = settigns.General.Foreground;
            this.lstSelector.BackColor = settigns.General.Background;

            this.hexViewer.ForeColor = settigns.General.Foreground;
            this.hexViewer.BackColor = settigns.General.Background;

            this.hexViewer.AddressWidth = Convert.ToInt32(settigns.HexView.Address);
            this.hexViewer.BlockWidth = Convert.ToInt32(settigns.HexView.Block);
            this.hexViewer.BytesPerLine = Convert.ToInt32(settigns.HexView.Bytes);
            this.hexViewer.HotTrack = settigns.HexView.HotTrack;
            this.hexViewer.UpperCase = settigns.HexView.UpperCase;
        }
    }
}

// EOF